@@ -12,10 +12,16 @@ Apache2::AuthEnv - Perl Authentication and Authorisation via Environment Variabl
 
  ### In httpd.conf or .htaccess: ################
  # Set the remote user and trigger the auth* stages
- AuthEnvVar %{REMOTE_ADDR}@%{SOME_ENV_VAR}
+ AuthEnvUser %{REMOTE_ADDR}@%{SOME_ENV_VAR}
+
+ # Also possible is setting the remote user from a list 
+ # of alternative environment variables or a default value.
+ AuthEnvUser %{HTTP_XX_USER|HTTP_YY_USER:anon}
 
  # Set extra environment variables.
- AuthEnvSet	HTTP_AE_SERVER	%{SERVER_ADDR}:%{SERVER_PORT}
+ AuthEnvSet	HTTP_AE_SERVER	%{SERVER_ADDR:unknown}:%{SERVER_PORT:unknown}
+ AuthEnvChange	HTTP_AE_SERVER	s/:/!/g
+ AuthEnvChange	HTTP_AE_SERVER	tr/a-z/A-Z/
 
  # Allow and Deny access based on environment.
  # The default is to deny access.
@@ -26,6 +32,10 @@ Apache2::AuthEnv - Perl Authentication and Authorisation via Environment Variabl
  AuthEnvDeny		%{REMOTE_ADDR}		192.168.2.3
  AuthEnvDenyMatch	%{HTTP_USER_AGENT}	Fedora
  AuthEnvAllow		%{SERVER_PORT} 80
+ AuthEnvAllowSplit	%{HTTP_MEMBEROF}   '\^' 'CN=....'
+
+ AuthEnvAllowAll
+ AuthEnvDenyAll
 
  AuthEnvDenial		UNAUTHORISED|UNAUTHORIZED|NOT_FOUND|FORBIDDEN
 
@@ -33,7 +43,7 @@ Apache2::AuthEnv - Perl Authentication and Authorisation via Environment Variabl
 
 B<Apache2::AuthEnv> allows you to promote a string composed of CGI
 environment variables to act as an authenticated user. The format is
-set via the AuthEnvVar command and the result is placed in the
+set via the AuthEnvUser command and the result is placed in the
 environment variable B<REMOTE_USER>.
 
 This module is for use only when another Apache module pre-authenticates
@@ -41,12 +51,12 @@ and pre-authorises a user but does not provide authentication nor
 authorisation controls within Apache.
 
 This module, once loaded, is triggered by the Apache directive
-I<AuthEnvVar> setting a format from the environment for the remote
+I<AuthEnvUser> setting a format from the environment for the remote
 user name. Authorisation is controlled by I<AuthEnvAllow*> and
 I<AuthEnvDeny*> directives. The default is to deny authorisation
 to everyone.
 
-  AuthEnvVar		%{HTTP_SSO_USER}@%{HTTP_SSO_ORG}
+  AuthEnvUser		%{HTTP_SSO_USER}@%{HTTP_SSO_ORG}
   AuthEnvAllowUser	fred@ORG
 
 Such a system is Computer Asscoiates' SiteMinder (c) Single Sign On
@@ -56,18 +66,26 @@ local web server. SiteMinder sets various environment variables
 including HTTP_SM_USER and HTTP_SM_AUTHDIRNAME. So a reasonable
 setting would be
 
-  AuthEnvVar		%{HTTP_SM_USER}@%{HTTP_SM_AUTHDIRNAME}
+  AuthEnvUser		%{HTTP_SM_USER}@%{HTTP_SM_AUTHDIRNAME}
   AuthEnvAllowUser	fred@ORG
 
 Another example is
-  AuthEnvVar		%{HTTP_UI_PRINCIPAL_NAME}
+  AuthEnvUser		%{HTTP_UI_PRINCIPAL_NAME}
   AuthEnvAllowUser	fred@ORG.org
   AuthEnvAllow		%{HTTP_UI_DEPARTMENT} sales
 
+Some systems may take authentication information from various sources
+and provide different environment variables for each source. So you can
+list alternative variables to use.
+  AuthEnvUser		%{HTTP_SOURCE1_NAME|HTTP_SOURCE2_NAME|HTTP_SOURCE3_NAME}
+
+If nothing matches then you can set a default value (say 'anon') via 
+  AuthEnvUser		%{HTTP_SOURCE_NAME|HTTP_SOURCE2_NAME:anon}
+
 For nested directives, configurations are inherited from one
-configuration file to the next. I<AuthEnvVar> directives overwrite each
+configuration file to the next. I<AuthEnvUser> directives overwrite each
 other as do collections of I<AuthEnvAllow*> rules. Each individual
-AuthEnvSet, unless overwriten, is inherited.
+AuthEnvSet and AuthEnvChange directive, unless overwriten, is inherited.
 
 The default denial code returned to the browser is FORBIDDEN.
 The directive I<AuthEnvDenial> can be used to change the return code.
@@ -75,6 +93,31 @@ For example,
 
   AuthEnvDenial		NOT_FOUND
 
+=head1 FORMAT
+
+The substitution format is composed of strings of characters and 
+variable substitutions starting with '%{' and ending in '}'.
+Substitutions are of the following formats:
+
+=over 2
+
+=item * %{ENVIRONMENT_VARIABLE_NAME},
+
+=item * %{ENVIRONMENT_VARIABLE_NAME1|ENVIRONMENT_VARIABLE_NAME2|....}
+
+=item * %{ENVIRONMENT_VARIABLE_NAME:default}.
+
+=back
+
+In the first case, the value of the environment variable is simply substituted. If a
+'|' separated list of variables is specified then each variable is
+checked in order, substituting the value of the first that is not empty.
+If no substitution succeeds and there is a default specified then that
+value is used instead.
+
+To use formats with spaces in the .htaccess file, enclose the format in
+double quotes.
+
 =head1 METHODS
 
 =over 4
@@ -94,11 +137,13 @@ directive in .htaccess and httpd.conf files.
 This is the method used as augument to the the PerlAuthzHandler
 directive in .htaccess and httpd.conf files.
 
+=back
+
 =head1 APACHE DIRECTIVES
 
 In the Apache configuration file httpd.conf, the module must be loaded
 
-=over 4
+=over 2
 
 PerlOptions +GlobalRequest
 
@@ -106,7 +151,9 @@ PerlModule Apache2::AuthEnv
 
 =back
 
-=item * AuthEnvVar <format>
+=over 4
+
+=item * AuthEnvUser <format>
 
 This turns on the authentication and authorisation stages and sets the
 format for the remote user name, which is filled in during
@@ -116,6 +163,11 @@ authentication.
 
 This sets the specified environment variable using the sepcified format.
 
+=item * AuthEnvSet <variable> <perl-substitution>
+
+This changes the specified environment variable according to the following
+Perl substitution. Modifications to REMOTE_USER are allowed.
+
 =item * AuthEnvAllowUser <user>
 
 =item * AuthEnvDenyUser <user>
@@ -153,11 +205,35 @@ repression and the others require exact matches.
 This is useful for environment variables that are really lists
 of values delimited with a specific value.
 
+Note that the <split> string is a regular expression and needs to be
+escaped appropiately; e.g. split on '\^' not on '^' as the latter just
+splits on the beginning of the string and is probably not what you want.
+
+=item * AuthEnvAllowFile <file>
+
+=item * AuthEnvDenyFile <file>
+
+These directives allow or deny, respectively,
+any users from the specified file.
+
+=item * AuthEnvAllowAll
+
+This directive allows any connection that hasn't been denied up to now.
+This is useful to allow all users to access the controlled area.
+
+=item * AuthEnvDenyAll
+
+This directive denies any connection that hasn't been allowed up to now.
+This is really the default action but included for completeness.
+It is useful when an area needs to be temporarily denied but the rest of the configuration needs to stay intact.
+
 =item * AuthEnvDenial	UNAUTHORISED|UNAUTHORIZED|NOT_FOUND|FORBIDDEN
 
 This directive sets the HTTP denial code returned to the
 browser if authorisation fails. The default is FORBIDDEN.
 
+=back
+
 =head1 AUTHOR
 
 Anthony R Fletcher arif@cpan.org
@@ -188,11 +264,13 @@ use strict;
 use warnings FATAL => 'all', NONFATAL => 'redefine';
 
 use vars qw($VERSION);
-$VERSION = '1.0';
+$VERSION = '1.3.1';
 
 use Carp;
 use Data::Dumper;
 
+use Safe;
+
 use ModPerl::Util;
 use Apache2::Module;
 use Apache2::Access ();
@@ -211,9 +289,13 @@ die "The module mod_perl 2.0 is required!" unless
 
 ###########################################################
 my @directives = (
+	{
+		name	=> 'AuthEnvUser',
+		errmsg	=> 'AuthEnvUser EnvVarFrormat',
+	},
 	{
 		name	=> 'AuthEnvVar',
-		errmsg	=> 'AuthEnvAllow EnvVarFrormat',
+		errmsg	=> 'AuthEnvVar EnvVarFrormat',
 	},
 	{
 		name		=> 'AuthEnvAllowUser',
@@ -265,11 +347,36 @@ my @directives = (
 		args_how	=> Apache2::Const::TAKE3,
 		errmsg		=> 'AuthEnvDenySplitMatch EnvVarFormat SplitRegEx RegEx',
 	},
+	{
+		name		=> 'AuthEnvAllowAll',
+		args_how	=> Apache2::Const::NO_ARGS,
+		errmsg		=> 'AuthEnvAllowAll',
+	},
+	{
+		name		=> 'AuthEnvDenyAll',
+		args_how	=> Apache2::Const::NO_ARGS,
+		errmsg		=> 'AuthEnvDenyAll',
+	},
+	{
+		name		=> 'AuthEnvAllowFile',
+		args_how	=> Apache2::Const::TAKE1,
+		errmsg		=> 'AuthEnvAllowFile <file>',
+	},
+	{
+		name		=> 'AuthEnvDenyFile',
+		args_how	=> Apache2::Const::TAKE1,
+		errmsg		=> 'AuthEnvDenyFile <file>',
+	},
 	{
 		name		=> 'AuthEnvSet',
 		args_how	=> Apache2::Const::TAKE2,
 		errmsg		=> 'AuthEnvSet EnvVar Format',
 	},
+	{
+		name		=> 'AuthEnvChange',
+		args_how	=> Apache2::Const::TAKE2,
+		errmsg		=> 'AuthEnvChange EnvVar <subsitution>'
+	},
 	{
 		name		=> 'AuthEnvDenial',
 		args_how	=> Apache2::Const::TAKE1,
@@ -298,7 +405,7 @@ sub new
 
 # Set the environment variable to use for authentication
 # and set the system to authenticate and authorise.
-sub AuthEnvVar
+sub AuthEnvUser
 {
 	my ($cfg, $parms, $fmt, @args) = @_;
 
@@ -312,21 +419,24 @@ sub AuthEnvVar
 	]);
 
 	# Check that the format contains something to expand.
-	unless ($fmt =~ /%{\w+}/)
+	unless ($fmt =~ /%\{.*\}/)
 	{
-		$r->server->log_error("AuthEnvVar format has no expansion! AuthEnv cancelled for ", $r->uri);
+		$r->server->log_error("AuthEnvUser format '$fmt' has no expansion! AuthEnv cancelled for ", $r->uri);
 		return Apache2::Const::HTTP_FORBIDDEN;
 	}
 
 	# Save value for user name format.
-	$cfg->{AuthEnvVar} = $fmt;
+	$cfg->{AuthEnvUser} = $fmt;
+
+	# Make sure the the user gets set later.
+	push @{$cfg->{set}}, ['set', 'REMOTE_USER', $fmt];
 
 	# Initialise the authorise rule list.
 	$cfg->{authorise} = ();
 
 	1;
 }
-
+sub AuthEnvVar { AuthEnvUser(@_); }
 
 # The @authorise array contains arrays of four elements:
 #	the environment format string,
@@ -334,6 +444,18 @@ sub AuthEnvVar
 #	if it's an exact (1) or a match rule (0).
 #	the string to compare/match it against.
 
+sub AuthEnvAllowAll
+{
+	my ($cfg, $parms) = @_;
+	push @{$cfg->{authorise}}, ['', 1, 1, undef, ''];
+}
+
+sub AuthEnvDenyAll
+{
+	my ($cfg, $parms) = @_;
+	push @{$cfg->{authorise}}, ['', 0, 1, undef, ''];
+}
+
 sub AuthEnvAllowUser
 {
 	my ($cfg, $parms, $user) = @_;
@@ -394,11 +516,69 @@ sub AuthEnvDenySplitMatch
 	push @{$cfg->{authorise}}, [$var, 0, 0, $split, $regex];
 }
 
+sub AuthEnvAllowFile
+{
+	my ($cfg, $parms, $file) = @_;
+
+	local *FILE;
+	unless (open (FILE, '<', $file))
+	{
+		warn "AuthEnv: Cannot read access allow file '$file' ($!).\n";
+		return;
+	}
+
+	local ($/) = undef; # slurp.
+	my $users = <FILE>;
+	$users =~ s/#.*$//gm;
+	
+	for my $user (split/\s+/, $users)
+	{
+		next unless ($user ne '');
+		push @{$cfg->{authorise}}, ['%{REMOTE_USER}', 1, 1, undef, $user];
+	}
+
+	close FILE;
+}
+
+sub AuthEnvDenyFile
+{
+	my ($cfg, $parms, $file) = @_;
+
+	local *FILE;
+	unless (open (FILE, '<', $file))
+	{
+		warn "AuthEnv: Cannot read access deny file '$file' ($!).\n";
+		warn "AuthEnv: Denying all!\n";
+
+		# deny all from this point; just in case.
+		&AuthEnvDenyAll($cfg, $parms);
+
+		return;
+	}
+
+	local ($/) = undef; # slurp.
+	my $users = <FILE>;
+	$users =~ s/#.*$//gm;
+	
+	for my $user (split /\s+/s, $users)
+	{
+		next unless ($user ne '');
+		push @{$cfg->{authorise}}, ['%{REMOTE_USER}', 0, 1, undef, $user];
+	}
+
+	close FILE;
+}
+
 sub AuthEnvSet
 {
 	my ($cfg, $parms, $var, $fmt) = @_;
+	push @{$cfg->{set}}, ['set', $var, $fmt];
+}
 
-	push @{$cfg->{set}}, [$var, $fmt];
+sub AuthEnvChange
+{
+	my ($cfg, $parms, $var, $change) = @_;
+	push @{$cfg->{set}}, ['change', $var, $change];
 }
 
 sub AuthEnvDenial
@@ -458,6 +638,45 @@ sub DIR_MERGE    { merge(@_) }
 sub SERVER_MERGE { merge(@_) }
 
 
+# Fill out a sub-format with the correct values.
+# Take a context ($r), a format of environment variables (with optional default) and 
+# a fail reference.
+# Return the value of the first environment variable that exists, or the default if specified
+# or '' and increament tehe failure variable reference.
+sub fillout
+{
+	my ($r, $fmt, $fail) = @_;
+
+	# check the format
+	#unless ($fmt =~ /^[\w\|]+(:\w+)?$/)
+	#{
+		# wrong format.
+ 		#$$fail++;
+		#$r->server->log_error("Invalid expansion '$fmt' at ", $r->uri);
+		#return '';
+	#}
+
+	#$r->server->log_error("Expanding '$fmt' at ", $r->uri);
+
+	# Isolate the default value.
+	my $default = ($fmt =~ s/:(\w*)$//) ? $1 : undef;
+
+	# Run though each environment valriable in turn.
+	for my $e (split(/\|/, $fmt))
+	{
+		# return value if it exists.
+		return $r->subprocess_env($e) if defined($r->subprocess_env($e));
+	}
+
+	# Otherwise return the default value.
+	return $default if defined $default;
+
+	# Failed.
+ 	$$fail++;
+
+	'';
+}
+
 ###########################################################
 
 # NB There is almost no environment to speak of at this time!
@@ -479,45 +698,77 @@ sub authenticate
 		$r->server->log_error("Wrong authentication Type ", $r->auth_type);
 		return Apache2::Const::HTTP_UNAUTHORIZED;
 	}
-	unless (defined $cfg->{AuthEnvVar})
+	unless (defined $cfg->{AuthEnvUser})
 	{
-		$r->server->log_error("AuthEnvVar not used! ", $r->auth_type);
+		$r->server->log_error("AuthEnvUser not used! ", $r->auth_type);
 		return Apache2::Const::HTTP_UNAUTHORIZED;
 	}
 
-	# Set AuthEnvVar value. Depreciated!
-	#my $AuthEnvVar = $r->dir_config('AuthEnvVar') || $cfg->{AuthEnvVar};
-
 	# Import CGI environment.
 	$r->subprocess_env unless $r->is_perl_option_enabled('SetupEnv');
 
-        # expand $AuthEnvVar format; fail if a variable doesn't
+        # expand $AuthEnvUser format; fail if a variable doesn't
         # not exist.
-	my $user = $cfg->{AuthEnvVar};
-	my $fail = 0; # count non-existant variables.
-	$user =~ s/%{(\w+)}/(defined($r->subprocess_env($1)) ? $r->subprocess_env($1) : $fail++)/gxe;
 
-	# Failure.
-	return Apache2::Const::HTTP_UNAUTHORIZED if $fail;
+	# Check that AuthEnvUser is set.
+	return Apache2::Const::HTTP_UNAUTHORIZED unless exists $cfg->{AuthEnvUser};
+
+	# Set the AE version environment.
+	$r->subprocess_env('HTTP_AE_VERSION', $VERSION);
+
+	# Set the environment and the REMOTE_USER along the way.
+	for my $s (@{$cfg->{set}})
+	{
+		my ($act, $v, $f) = @$s;
+
+		# Set an environment variable.
+		if ($act eq 'set')
+		{
+			my $fail = 0; # count non-existant variables.
+
+			#$r->server->log_error($r->uri, ": change '$f'");
+
+			$f =~ s/%\{([^\}]+)\}/&fillout($r, $1, \$fail)/gxe;
+
+			# something wasn't defined.
+			return Apache2::Const::HTTP_UNAUTHORIZED if $fail;
+
+			$r->subprocess_env($v, $f);
+		}
+		# Change an environment variable.
+		elsif ($act eq 'change')
+		{
+			my $val = $r->subprocess_env($v);
+
+			# Run the modification in a safe environment.
+			my $cpt = new Safe;
+			${$cpt->varglob('val')} = $val;
+			$cpt->reval("\$val =~ $f");
+
+			if ($@)
+			{
+				# failure to run.
+				$r->server->log_error("change '$f' failed ($@) ", $r->uri);
+				return Apache2::Const::HTTP_UNAUTHORIZED;
+			}
+			else
+			{
+				# success.
+				$r->subprocess_env($v,${$cpt->varglob('val')});
+			}
+		}
+
+		# Set the authenticated user as we go.
+		$r->user($r->subprocess_env('REMOTE_USER'))
+				if ($v eq 'REMOTE_USER');
+	}
+
+	# Check that the user is real.
+	my $user = $r->user();
 	return Apache2::Const::HTTP_UNAUTHORIZED unless defined $user;
 	return Apache2::Const::HTTP_UNAUTHORIZED if ($user eq '');
 
-	# Success! User authenticated
-
-	# Set REMOTE_USER, etc..
-	$r->subprocess_env('REMOTE_USER', $user);
-	$r->user($user);
-
-	# Set the environment.
-	$r->subprocess_env('HTTP_AE_VERSION', $VERSION);
-	for my $s (@{$cfg->{set}})
-	{
-		my ($v, $f) = @$s;
-		$f =~ s/%{(\w+)}/$r->subprocess_env($1)/gxe;
-		$r->subprocess_env($v, $f);
-	}
-
-	# succeed.
+	# Success.
 	return Apache2::Const::OK;
 }
 
@@ -536,7 +787,7 @@ sub allowed
 		my $fail = 0; # count non-existant variables.
 
 		# Substitute.
-		$val =~ s/%{(\w+)}/(defined($r->subprocess_env($1)) ? $r->subprocess_env($1) : $fail++)/gxe;
+		$val =~ s/%\{([^\}]+)\}/&fillout($r, $1, \$fail)/gxe;
 
 		# Fail if this contains a non-existant environment variable.
 		return 0 if $fail;
@@ -557,7 +808,12 @@ sub allowed
 				: ($v =~ m/$regex/);
 
 
-			return $allow if $match;
+			#return $allow if $match;
+			if ($match)
+			{
+				#debug $r, "match '$v' against '$regex' returns '$allow'\n";
+				return $allow;
+			}
 		}
 	}
 
